import hashlib
from typing import Literal, Optional

HashName = Literal[
    'md5',
    'sha1',
    'sha224',
    'sha256',
    'sha384',
    'sha512',
    'blake2b',
    'blake2s',
    'sha3_224',
    'sha3_256',
    'sha3_384',
    'sha3_512',
    'shake_128',
    'shake_256',
]


def hash_file(path: str, hash_name: HashName = 'sha256', chunk_size: Optional[int] = None) -> str:
    """Return the SHA-256 hash of the file at the given path."""
    try:
        hash_fn = getattr(hashlib, hash_name)
    except AttributeError:
        raise ValueError(f"Invalid hash name: {hash_name}")

    with open(path, 'rb') as f:
        if chunk_size:
            hash = hash_fn()
            for chunk in iter(lambda: f.read(chunk_size), b''):
                hash.update(chunk)
        else:
            hash = hash_fn(f.read())
        return hash.hexdigest()



#!/usr/bin/env python
import shlex
from functools import partial
from json import loads
from subprocess import check_call, check_output, CalledProcessError, DEVNULL, Popen, PIPE, STDOUT
from sys import stderr
from typing import Optional, List, Tuple, Callable, Union, Dict

err = partial(print, file=stderr)
Log = Optional[Callable[..., None]]
Arg = Union[None, str, int, List['Arg'], Tuple['Arg']]


def silent(*args) -> None:
    pass


def flatten(args: Union[List[Arg], Tuple[Arg, ...]]) -> Tuple:
    """Recursively flatten a nested list of arguments."""
    if isinstance(args, list) or isinstance(args, tuple):
        return tuple(
            a
            for arg in args
            for a in flatten(arg)
        )
    else:
        return (args,)


def parse_cmd(cmd: Tuple[Arg, ...]) -> List[str]:
    """Flatten and stringify a command."""
    return [
        str(arg)
        for arg in flatten(cmd)
        if arg is not None
    ]


def lines(
    *cmd: Arg,
    keep_trailing_newline: bool = False,
    dry_run: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> Optional[List[str]]:
    """Return the lines written to stdout by a command."""
    out = output(*cmd, dry_run=dry_run, err_ok=err_ok, **kwargs)
    if err_ok and out is None:
        return None

    lines = [
        line.rstrip('\n')
        for line in
        out.decode().split('\n')
    ]

    if not keep_trailing_newline and lines and not lines[-1]:
        lines = lines[:-1]

    return lines


def line(
    *cmd: Arg,
    empty_ok: bool = False,
    err_ok: bool = False,
    **kwargs,
) -> Optional[str]:
    """Run a command, verify that it returns a single line of output, and return that line."""
    _lines = lines(*cmd, err_ok=err_ok, **kwargs)
    if (empty_ok or err_ok) and not _lines:
        return None
    elif len(_lines) == 1:
        return _lines[0]
    else:
        raise ValueError(f'Expected 1 line, found {len(_lines)}:\n\t%s' % '\n\t'.join(_lines))


ELIDED = '****'
Elides = Union[None, str, List[str]]


def mk_cmd_str(cmd: List[str], elide: Elides = None) -> str:
    """Convert a command to a string, optionally eliding known sensitive values."""
    cmd_str = shlex.join(cmd)
    if elide:
        if isinstance(elide, str):
            elide = [elide]
        for s in elide:
            cmd_str = cmd_str.replace(s, ELIDED)
    return cmd_str


def run(
    *cmd: Arg,
    dry_run: bool = False,
    elide: Elides = None,
    log: Log = err,
    **kwargs,
) -> None:
    """Convenience wrapper for ``subprocess.check_call``."""
    cmd = parse_cmd(cmd)
    cmd_str = mk_cmd_str(cmd, elide)
    if dry_run:
        if log:
            log(f'Would run: {cmd_str}')
    else:
        if log:
            log(f'Running: {cmd_str}')
        check_call(cmd, **kwargs)


sh = run


def interleaved_output(cmd: List[str], err_ok: bool = False) -> bytes:
    try:
        proc = Popen(cmd, stdout=PIPE, stderr=STDOUT)

        # Stream and capture the output in real-time
        output = b''
        for line in proc.stdout:
            output += line

        proc.wait()
        if proc.returncode != 0:
            raise CalledProcessError(proc.returncode, cmd, output=output)
        else:
            return output
    except CalledProcessError as e:
        if err_ok:
            return e.output
        else:
            err(e.output.decode())
            raise e


def output(
    *cmd: Arg,
    dry_run: bool = False,
    both: bool = False,
    err_ok: bool = False,
    elide: Elides = None,
    log: Log = err,
    **kwargs,
) -> Optional[bytes]:
    """Convenience wrapper for ``subprocess.check_output``.

    By default, logs commands to `err` (stderr) before running (pass `log=None` to disable).

    If `err_ok=True`, exceptions will be caught, and `None` returned.

    If `both=True`, stdout and stderr will both be captured (interleaved), and returned.
    """

    cmd = parse_cmd(cmd)
    cmd_str = mk_cmd_str(cmd, elide)
    if dry_run:
        if log:
            log(f'Would run: {cmd_str}')
        return None
    else:
        if log:
            log(f'Running: {cmd_str}')
        if both:
            return interleaved_output(cmd, err_ok=err_ok)
        else:
            try:
                return check_output(cmd, **kwargs)
            except CalledProcessError as e:
                if err_ok:
                    return None
                else:
                    raise e


def json(
    *cmd: Arg,
    dry_run: bool = False,
    **kwargs,
) -> Union[None, List, Dict, str, int, float, bool]:
    """Run a command, parse the output as JSON, and return the parsed object."""
    out = output(*cmd, dry_run=dry_run, **kwargs)
    if out is None:
        return None
    return loads(out.decode())


def check(
    *cmd: Arg,
    stdout=DEVNULL,
    stderr=DEVNULL,
    **kwargs,
):
    """Run a command, return True iff it runs successfully (i.e. exits with code 0)."""
    try:
        run(*cmd, stdout=stdout, stderr=stderr, **kwargs)
        return True
    except CalledProcessError:
        return False


# Omit "json", to avoid colliding with stdlib
__all__ = [ 'check', 'err', 'flatten', 'interleaved_output', 'line', 'lines', 'output', 'run', 'sh', 'silent', ]



from contextlib import contextmanager
from os import mkdir
from os.path import join
from tempfile import TemporaryDirectory


@contextmanager
def tmpdir(name: str = None, dir: str = None):
    """``contextmanager`` for creating a temporary directory with an optional ``name``"""
    if name:
        with TemporaryDirectory(dir=dir) as tmpdir:
            dir = join(tmpdir, name)
            mkdir(dir)
            yield dir
    else:
        yield TemporaryDirectory(dir=dir)



# # Diff DFs
# Compute various diffs between two Pandas DataFrames

from IPython.display import HTML
from pandas import concat, Index, IndexSlice, isna, MultiIndex
from re import sub

def neq(l, r):
    return l != r and not (isna(l) and isna(r))


def blank_row(row, fill=''):
    l,r = row.iloc[:]
    if neq(l,r):
        return row
    else:
        row.loc[:] = (fill,fill)
        return row


def blank_eqs(df, fill=''):
    return df.apply(blank_row, axis=1, fill=fill)


class Diff:
    def __init__(self, l, r, **join_kwargs):
        self.l = l
        self.r = r

        l_col_set = self.l_col_set = set(l.columns)
        r_col_set = self.r_col_set = set(r.columns)
        l_only_col_names = self.l_only_col_names = [ c for c in l.columns if c not in r_col_set ]
        r_only_col_names = self.r_only_col_names = [ c for c in r.columns if c not in l_col_set ]
        lr_col_names = self.lr_col_names = Index([ c for c in l.columns if c in r_col_set ])

        l_only_cols = self.l_only_cols = l[l_only_col_names]
        r_only_cols = self.r_only_cols = r[r_only_col_names]

        l_row_set = self.l_row_set = set(l.index)
        r_row_set = self.r_row_set = set(r.index)
        l_only_row_names = self.l_only_row_names = [ c for c in l.index if c not in r_row_set ]
        r_only_row_names = self.r_only_row_names = [ c for c in r.index if c not in l_row_set ]
        lr_index = self.lr_index = Index([ c for c in l.index if c in r_row_set ])

        l_only_rows = self.l_only_rows = l.loc[l_only_row_names]
        r_only_rows = self.r_only_rows = r.loc[r_only_row_names]        

        l_shared = self.l_shared = l.loc[lr_index, lr_col_names]
        r_shared = self.r_shared = r.loc[lr_index, lr_col_names]

        groups = self.groups = {
            c: concat([l_shared[c].rename('l'),r_shared[c].rename('r')], axis=1)
            for c in lr_col_names
        }

        l_suffix, r_suffix = join_kwargs.get('suffixes', ('_x','_y'))
        suffixes_regex = f'(?:{l_suffix}|{r_suffix})$'
        merged = self.merged = l.merge(r, left_index=True, right_index=True, **join_kwargs)
        grouped = self.grouped = merged.groupby(lambda c: sub(suffixes_regex,'',c), axis=1)

        cols = self.cols = lr_col_names
        midx = MultiIndex.from_tuples([ (c, side) for c in cols for side in ['l','r'] ])

        # Build DF with top-level columns and {l,r} sub-columns with both sides' values
        df = self.df = concat(
            [ 
                (l_shared[c] if sfx == 'l' else r_shared[c]) \
                .rename((c, sfx))
                for c, sfx in midx
            ],
            axis=1,
        )
        df.columns.names = ['col','side']
        midx = self.midx = df.columns

        neqs = self.neqs = ((l_shared != r_shared) & (l_shared.notnull() | r_shared.notnull()))

        row_counts = self.row_counts = neqs.sum(axis=1)
        rows_changed = self.rows_changed = self.rows_changed = row_counts>0
        changed_rows = self.changed_rows = lr_index[rows_changed]        

        col_counts = self.col_counts = neqs.sum()
        cols_changed = self.cols_changed = col_counts>0
        changed_cols = self.changed_cols = lr_col_names[cols_changed]

        changed = self.changed = df.loc[changed_rows, IndexSlice[changed_cols,:]]

        clean = self.clean =             self             .changed             .groupby(level=0,axis=1)             .apply(blank_eqs, fill='')
    
    def _repr_html_(self): return self.clean._repr_html_()
    
    def __getattr__(self, k):
        if hasattr(self.changed, k):
            return getattr(self.changed, k)
        raise AttributeError(k)

    def __getitem__(self, k):
        if k in self.changed:
            return self.changed[k]
        raise AttributeError(k)



import json
from functools import partial
from io import TextIOWrapper
from os import environ as env, makedirs
from os.path import exists, join, relpath
from sys import stderr
from typing import Union, Literal, Tuple, Callable, Optional, Dict, List

from IPython.display import Image

import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.validators.scatter.marker import SymbolValidator
symbols = SymbolValidator().values[2::12]

from utz import err


# Env vars to fall back to, for various plot kwargs
PLOT_SHOW_VAR = 'UTZ_PLOT_SHOW'
PLOT_DIR_VAR = 'UTZ_PLOT_DIR'
PLOT_MARGIN_VAR = 'UTZ_PLOT_MARGIN'
PLOT_BG_VAR = 'UTZ_PLOT_BG'
PLOT_GRID_VAR = 'UTZ_PLOT_GRID'
PLOT_SUBTITLE_SIZE_VAR = 'UTZ_PLOT_SUBTITLE_SIZE'
PLOT_HTML_VAR = 'UTZ_PLOT_HTML'

DEFAULT_BG = 'white'
DEFAULT_SHOW = 'html'
DEFAULT_GRID = '#ccc'
DEFAULT_SUBTITLE_SIZE = '0.8em'


class Unset:
    pass


_Unset = type(Unset)


Title = Union[str, List[str], None]


def title(s: Title, subtitle_size: Title = None) -> Optional[str]:
    """Convert a list of strings into a `<br>`-joined plot title, with optional font-size styling.

    If `subtitle_size` is a `str`, all elements of `s` after the first (subtitles) will be given that size.
    If it's a `list[str]`, the sizes will be applied to each element of `s`, with the last "size" repeated out to the
    length of the `s` list."""
    if isinstance(s, list):
        if subtitle_size is None:
            subtitle_size = env.get(PLOT_SUBTITLE_SIZE_VAR, DEFAULT_SUBTITLE_SIZE)

        title0, *subtitles = s
        if isinstance(subtitle_size, list) and subtitle_size:
            sizes = [
                subtitle_size[i] if i < len(subtitle_size) else subtitle_size[-1]
                for i in range(len(s))
            ]
            title_size, *subtitle_sizes = sizes
            title0 = f'<span style="font-size:{title_size}">{title0}</span>'
            subtitles = [
                f'<span style="font-size:{subtitle_size}">{subtitle}</span>'
                for subtitle_size, subtitle in zip(subtitle_sizes, subtitles)
            ]
        elif isinstance(subtitle_size, str):
            subtitles = [
                f'<span style="font-size:{subtitle_size}">{subtitle}</span>'
                for subtitle in subtitles
            ]
        return '<br>'.join([title0, *subtitles ])
    else:
        return s


make_title = title


def save(
    fig: Figure,
    name: str,
    title: Title = None,
    **kwargs
):
    """Wrapper around ``plot``, enforces non-``None`` ``name`` and allows ``name`` then ``title`` as positional args."""
    if not name:
        raise ValueError("`name` must be nonempty")
    return plot(fig, name=name, title=title, **kwargs)


def plot(
    fig: Figure,
    title: Title = None,
    name: Optional[str] = None,
    *,
    bg: Union[str, None, _Unset] = Unset,
    subtitle_size: Optional[str] = None,
    hoverx: bool = False,
    hovertemplate: Title = None,
    png_title: bool = True,
    yrange: Optional[str] = 'tozero',
    legend: Union[bool, Dict, Literal['reversed'], None] = None,
    bottom_legend: Union[bool, Literal['all']] = False,
    pretty: bool = False,
    margin: Union[None, int, Dict, _Unset] = Unset,
    dir: Optional[str] = None,
    w: Optional[int] = None,
    h: Optional[int] = None,
    png: Union[Dict, int, Tuple[int], Tuple[int, int], None] = None,
    xtitle: Title = None,
    ytitle: Title = None,
    ltitle: Title = None,
    grid: Union[str, None, _Unset] = Unset,
    xgrid: Union[str, None, _Unset] = Unset,
    ygrid: Union[str, None, _Unset] = Unset,
    x: Union[Dict, str, None] = None,
    y: Union[Dict, str, None] = None,
    log: Union[TextIOWrapper, bool, None] = stderr,
    show: Union[Literal['png', 'file', 'html'], bool, None] = None,
    zerolines: Union[bool, Literal['x', 'y'], None] = True,
    html: Union[Dict, bool, None, _Unset] = Unset,
    title_suffix: Optional[str] = None,
    **layout,
):
    """Plotly wrapper with convenience kwargs for common configurations.

    Args:
        fig: Plotly figure
        title: Title string or list of strings (tail will be converted to "subtitle" lines)
        name: "Stem" for output file paths
        bg: Background color, falls back to $UTZ_PLOT_BG, defaults to `DEFAULT_BG` ("white")
        subtitle_size: Subtitle font size; falls back to $UTZ_PLOT_SUBTITLE_SIZE, defaults to "0.8em"
        hoverx: Alias for `hovermode="x"`
        hovertemplate: Hover template; `list[str]` will be `<br>.join()`'d.
        png_title: Whether to include the title in PNG (sometimes nice to disable this, e.g. for PNGs that will be embedded in Markdown that includes the title).
        yrange: Y-axis range mode
        legend: Legend configuration
        bottom_legend: Orient the legend horizontally along the bottom; True ⇒ JSON/PNG outputs only, 'all' ⇒ also in returned Figure
        pretty: Pretty-print JSON output
        margin: Margin size or dict of sizes (falls back to $UTZ_PLOT_MARGIN)
        dir: Output directory (falls back to $UTZ_PLOT_DIR)
        w: Width; applied to Figure and any resulting PNG output
        h: Height; applied to Figure and any resulting PNG output
        png: layout options specific to saved/static representations of the plot (omitted from JSON, included in PNG/HTML); "w" and "h" are short-hands for "width" and "height", a pair of ints will be interpreted as (w, h), and a single int or (int,) tuple will be interpreted as width.
        xtitle: X-axis title; string or list of strings (tail will be converted to "subtitle" lines)
        ytitle: Y-axis title; string or list of strings (tail will be converted to "subtitle" lines)
        ltitle: Legend title; string or list of strings (tail will be converted to "subtitle" lines)
        grid: Grid color (x- and y-axes)
        xgrid: X-axis grid color
        ygrid: Y-axis grid color
        x: X-axis configuration (passed to `update_xaxes`)
        y: Y-axis configuration (passed to `update_yaxes`)
        log: Whether/where to print log info (e.g. about files written); defaults to stderr
        show: Format to return the Figure in: 'png' ⇒ return PNG `Image`, 'file' ⇒ return `Image(filename=…)` pointing to output `.png`, 'html' ⇒ return default Plotly Figure HTML, False ⇒ None (don't show figure, if `plot(…)` call is last expression in a notebook cell). Falls back to $UTZ_PLOT_SHOW, defaults to 'html'.
        zerolines: Whether to show zero lines; 'x' ⇒ only on x-axis, 'y' ⇒ only on y-axis, True (default) ⇒ on both axes
        title_suffix: When saving static plots (e.g. PNG), and this is set, two versions will be saved: one including the title, and one not. This string will be appended to the filename stem of the former. Assumes the default is the "untitled" version. Useful when embedding in contexts where it looks nicer to render the title in surrounding Markdown/HTML.
        html: When truthy, save plot as HTML. Value is interpreted as kwargs for Figure.to_html; falls back to json.loads($UTZ_PLOT_HTML).
    """
    mk_title = partial(make_title, subtitle_size=subtitle_size)
    ltitle = mk_title(ltitle)
    layout['legend_title'] = layout.get('legend_title', ltitle or '')

    if w is not None:
        layout['width'] = w
    if h is not None:
        layout['height'] = h

    if bg is Unset:
        bg = env.get(PLOT_BG_VAR)
        if bg is None:
            bg = DEFAULT_BG

    if bg:
        layout['plot_bgcolor'] = bg
        layout['paper_bgcolor'] = bg

    if isinstance(hovertemplate, list):
        hovertemplate = '<br>'.join(hovertemplate)
    if hoverx:
        layout['hovermode'] = 'x'
        fig.update_traces(hovertemplate=hovertemplate)
    elif hovertemplate:
        fig.update_traces(hovertemplate=hovertemplate)

    if yrange:
        layout['yaxis_rangemode'] = yrange

    if legend is False:
        layout['showlegend'] = False
    elif legend == 'reversed':
        layout['legend'] = dict(traceorder='reversed')
    elif isinstance(legend, dict):
        layout['legend'] = legend

    bottom_legend_kwargs = dict(
        orientation='h',
        x=0.5,
        xanchor='center',
        yanchor='top',
    ) if bottom_legend else {}
    if bottom_legend == 'all':
        if 'legend' not in layout:
            layout['legend'] = {}
        layout['legend'].update(**bottom_legend_kwargs)

    if grid is Unset:
        grid = env.get(PLOT_GRID_VAR)
        if grid is None:
            grid = DEFAULT_GRID

    def update_axis(
        xy: Union[str, Dict, None],
        xy_name: str,
        update_axes: Callable,
        xy_title: Title,
        xy_grid: Union[str, None, _Unset],
    ):
        if isinstance(xy, str):
            update_axes(title_text=xy)
        elif isinstance(xy, dict):
            if 'title' in xy:
                xy['title_text'] = xy['title']
                del xy['title']
            update_axes(**xy)

        if xy_grid is not Unset:
            update_axes(gridcolor=xy_grid)
        elif grid:
            update_axes(gridcolor=grid)

        if zerolines == xy_name or zerolines is True:
            update_axes(
                zeroline=True,
                zerolinecolor=xy_grid or grid,
                zerolinewidth=1,
            )

        xy_title = mk_title(xy_title)
        if xy_title:
            update_axes(title=dict(text=xy_title))

    update_axis(x, 'x', fig.update_xaxes, xtitle, xgrid)
    update_axis(y, 'y', fig.update_yaxes, ytitle, ygrid)

    title = mk_title(title)
    title_layout = dict(title_text=title, title_x=0.5) if title else {}
    if png_title:
        layout.update(title_layout)
    fig.update_layout(**layout)

    saved = go.Figure(fig)
    if not png_title:
        # only need to do this if it wasn't already done above
        fig.update_layout(**title_layout)

    if bottom_legend is True:
        saved.update_layout(
            legend=bottom_legend_kwargs,
        )

    if margin is Unset:
        margin = env.get(PLOT_MARGIN_VAR)
        if margin:
            margin = json.loads(margin)
    if isinstance(margin, int):
        margin = { k: margin for k in 'trbl' }
    if margin:
        saved.update_layout(margin=margin)

    if log is True:
        log = err
    elif log:
        log = partial(print, file=log)
    else:
        log = lambda _: ()

    png_path = None
    if name:
        if dir is None:
            dir = env.get(PLOT_DIR_VAR)
        if dir:
            if not exists(dir):
                makedirs(dir, exist_ok=True)
        else:
            dir = '.'
        json_path = join(dir, f'{name}.json')
        saved.write_json(json_path, pretty=pretty)
        png_path = join(dir, f'{name}.png')
        log(f"Wrote plot JSON to {relpath(json_path)}")
        if png:
            # Convert `int | Tuple[int] | Tuple[int, int]` to `dict`
            if isinstance(png, tuple):
                if len(png) == 1:
                    png = dict(w=png[0])
                elif len(png) == 2:
                    w, h = png
                    png = dict(w=w, h=h)
                else:
                    raise ValueError(f"Invalid PNG tuple: {png}")
            elif isinstance(png, int):
                png = dict(w=png)

            # `png` is a `dict`
            if 'w' in png:
                saved.update_layout(width=png['w'])
                del png['w']
            if 'h' in png:
                saved.update_layout(height=png['h'])
                del png['h']
            if 'title' in png and isinstance(png['title'], str):
                saved.update_layout(title_text=png['title'])
                del png['title']

            saved.update_layout(**png)

        if title_suffix:
            no_title = go.Figure(saved)
            no_title.update_layout(title_text='')
            title_path = join(dir, f'{name}{title_suffix}.png')
            no_title.write_image(png_path)
            log(f"Wrote plot image to {relpath(png_path)}")
            saved.write_image(title_path)
            log(f"Wrote plot image to {relpath(title_path)}")
        else:
            saved.write_image(png_path)
            log(f"Wrote plot image to {relpath(png_path)}")

        if html is Unset:
            html = env.get(PLOT_HTML_VAR)
            if html:
                html = json.loads(html)
        if html:
            html_path = join(dir, f'{name}.html')
            html_kwargs = dict(include_plotlyjs='cdn')
            if isinstance(html, dict):
                html_kwargs.update(html)
            saved.write_html(html_path, **html_kwargs)
            log(f"Wrote HTML to {relpath(html_path)}: {html}")

    if show is None:
        show = env.get(PLOT_SHOW_VAR)
        if show is None:
            show = DEFAULT_SHOW
    if show == 'file':
        if png_path:
            return Image(filename=png_path)
        else:
            raise ValueError("No `name` passed, for `show='file'`")
    elif show == 'png':
        return Image(fig.to_image())
    elif show is None or show is False:
        return
    else:
        return fig



import dataclasses
from contextlib import contextmanager
from dataclasses import replace
from inspect import getfullargspec, getmembers, isfunction
from typing import Any, Callable, Iterable, Optional, Type, TypeVar, Union, Dict, Tuple

Case = TypeVar('Case')  # Dataclass type
IdFmtField = Callable[[Any], str]
IdFmts = Dict[str, IdFmtField]
IdFmtsInput = Dict[Union[str, Tuple[str]], IdFmtField]


def default_field_fmt(val: Any) -> str:
    """Default field-formatting function for ``Case`` IDs."""
    return f"{val}"


def normalize(id_fmts: IdFmtsInput) -> IdFmts:
    return {
        key: fn
        for keys, fn in id_fmts.items()
        for key in (keys if isinstance(keys, tuple) else [keys])
    }


def parametrize(
    *cases: Union[Case, Iterable[Case]],
    delim: str = "-",
    id_fmts: Optional[IdFmtsInput] = None,
    **sweeps,
):
    """"Parametrize" [sic] a test function with a list of test-"case"s (instances of a dataclass).

    Fields, `@property``s, and methods of each ``case`` may be passed as keyword arguments to the
    test function, if it declares arguments with matching names. Test functions can also declare an
     argument named ``case`` to receive the full ``Case`` object.

    Test-case "ID"s (used by pytest for display and lookup purposes, for each test case) are taken
    from each ``Case``'s ```id`` attribute, if present. Otherwise, an ID is generated from the
    dataclass-fields whose values differ among ``cases``.

    ``**sweeps`` supports parameter sweeps, where ``cases`` is "Cartesian product"-ed against a
     series of ``(str, list[Any])`` pairs (where the "key" is a dataclass-field name).

    Adapted/Extended from https://github.com/single-cell-data/TileDB-SOMA/blob/1.14.2/apis/python/tests/parametrize_cases.py.
    """
    # Flatten the ``cases`` varargs. Sometimes it's convenient to pass a Generator as a single arg
    # to ``parametrize`` (e.g. to construct parameter sweeps with a loop), or otherwise build one
    # or more ``Sequence``s of ``Case``s to pass to ``parametrize``.
    if len(cases) == 1 and isinstance(cases[0], list):
        # Special-case, to save ``O(N)`` redundant traversals when N ``sweeps`` are provided (see
        # recursive ``sweeps`` unroll block below)
        cases = cases[0]
    else:
        cases = [
            case
            for arg in cases
            for case in (arg if isinstance(arg, Iterable) else [arg])
        ]

    if sweeps:
        # Pop first ``(key,vals)`` pair
        key = next(iter(sweeps.keys()))
        vals = sweeps.pop(key)
        # Sweep over ``vals``, recurse until no ``sweeps`` remain
        return parametrize(
            [
                replace(case, **{ key: val })
                for case in cases
                for val in vals
            ],
            delim=delim,
            id_fmts=id_fmts,
            **sweeps,
        )

    # Verify and extract a single ``class`` type from ``cases``.
    cls = cases[0].__class__
    for case in cases:
        if case.__class__ is not cls:
            raise ValueError(
                f"Expected all cases to be of type {cls}, but found {case.__class__}"
            )

    # Dicts of ``@property``s, methods, and ``@dataclass`` fields.
    props = {
        name: prop
        for name, prop
        in getmembers(cls, lambda o: isinstance(o, property))
    }
    methods = {
        name: method
        for name, method
        in getmembers(cls, isfunction)
        if not name.startswith('_')
    }
    fields = { f.name: f for f in dataclasses.fields(cls) }

    # Find fields that differ among the ``cases``; used for auto-generating test-case IDs
    differing_fields = {}
    if cases:
        case0, *rest = cases
        for name, field in fields.items():
            val0 = getattr(case0, name)
            for case in rest:
                val = getattr(case, name)
                if val != val0:
                    differing_fields[name] = field
                    break

    id_fmts = normalize(id_fmts) if id_fmts else {}
    _id_fmts = normalize(getattr(cls, '_id_fmts', {}))
    id_fmts = {
        name: id_fmts.get(name, _id_fmts.get(name, default_field_fmt))
        for name in differing_fields.keys()
    }

    def wrapper(fn):
        """Parameterize a test ``fn``, converting the ``cases`` above to "ID"s, key,
        and value lists that Pytest expects.

        The wrapped function's argument list can be any subset of the "Case" dataclass' fields,
        ``@property``s, and methods. Fields not "declared" by the test function are not passed.
        """
        def get_case_id(case: Case) -> str:
            """Generate a test-case ID from a "Case" object.

            Hyphen-delimited string reprs of all fields that differ among ``cases``."""
            case_id = getattr(case, 'id', None)
            if case_id:
                return case_id
            id_vals = []
            for name, field in differing_fields.items():
                val = getattr(case, name)
                fmt_fn = id_fmts[name]
                val_str = fmt_fn(val)
                id_vals.append(val_str)
            return delim.join(id_vals)

        # Test-case IDs, used by Pytest for display and lookup purposes
        ids = [ get_case_id(case) for case in cases ]

        # Full list of fields that can be passed to the test function
        # "case" is the dataclass instance itself
        cls_names: list[str] = [
            *fields.keys(),
            *props.keys(),
            *methods.keys(),
            'case',
        ]

        # Intersect the test function's argument names with the eligible field names above
        spec = getfullargspec(fn)
        fn_arg_names: list[str] = [
            arg
            for arg in spec.args
            if arg in cls_names
        ]

        # Dicts of keys and values, one per ``case``, for the keys (kwargs) expected by ``fn``
        # cases_kvs: list[dict[str, Any]] = []
        values_lists: list[list[Any]] = []
        for case in cases:
            values = []
            for name in fn_arg_names:
                if name == 'case':
                    values.append(case)
                elif name in fields:
                    values.append(getattr(case, name))
                elif name in props:
                    values.append(props[name].fget(case))
                elif name in methods:
                    values.append(getattr(case, name))
            values_lists.append(values)

        import pytest

        # Delegate to PyTest `parametrize`
        return pytest.mark.parametrize(
            fn_arg_names,  # List of kwarg ("key") names that match eligible ``Case`` class members
            values_lists,  # List of arg-value lists, one per test ``case``
            ids=ids,       # List of test-case names
        )(fn)

    return wrapper


@contextmanager
def raises(
    exc_type: Type[Exception],
    match: Union[str, list[str], None] = None,
    exact: Optional[bool] = None,
    *args,
    **kwargs,
):
    """``pytest.raises`` wrapper, 2nd arg can be ``match`` regex or ``list`` of exact-match candidates."""
    import pytest
    if exact:
        if isinstance(match, str):
            match = [match]
    if isinstance(match, list):
        msgs = match
        with pytest.raises(exc_type) as exc:
            yield
        expected = str(exc.value)
        assert expected in msgs, f'Expected "{expected}" to be in {msgs}'
    else:
        with pytest.raises(exc_type, *args, match=match, **kwargs):
            yield



# # Context Manager utilities

from contextlib import AbstractContextManager, contextmanager, nullcontext, suppress
from typing import ContextManager, Sequence, Union


def contexts(*ctxs: Union[ContextManager, Sequence[ContextManager]]) -> ContextManager:
    """Compose context managers."""
    ctxs = [
        ctx
        for _ctxs in ctxs
        for ctx in (
            _ctxs
            if isinstance(_ctxs, Sequence)
            else [_ctxs]
        )
    ]

    @contextmanager
    def fn(ctxs):
        if not ctxs:
            yield
        else:
            [ ctx, *rest ] = ctxs
            with ctx as v, fn(rest) as vs:
                yield [ v, *(vs if vs else []) ]

    return fn(ctxs)


ctxs = contexts


class catch(AbstractContextManager):
    """``contextmanager`` that catches+verifies ``Exception``s"""
    def __init__(self, *excs):
        self.excs = excs

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            if len(self.excs) == 1:
                raise AssertionError(f'No {self.excs[0].__name__} was thrown')
            else:
                raise AssertionError(f'None of {",".join([e.__name__ for e in self.excs])} were thrown')
        
        if not [ isinstance(exc_value, exc) for exc in self.excs ]:
            raise exc_value

        return True


# ## `no`: context manager for verifying `NameError`s (undefined variable names)
no = catch(NameError)




from datetime import datetime as dt
from sys import stderr

from .o import o


class now:
    try:
        from pytz import UTC
        pytz = True
        def tz(d): return d.astimezone(now.UTC)
    except ImportError:
        pytz = False
        UTC = None
        def tz(d): return d

    EPOCH = dt(1970,1,1).replace(tzinfo=UTC)
    FMTS = o(
        iso='%Y-%m-%dT%H:%M:%SZ',
        short='%Y%m%dT%H%M%SZ',
        micro='%Y-%m-%dT%H:%M:%S.%fZ',
    )

    @property
    def fmts(self):
        return self.FMTS

    def __init__(self, fmt=None, d=None):
        FMTS = self.FMTS
        if fmt is None:
            fmt = FMTS.iso
        elif isinstance(fmt, str):
            if hasattr(FMTS, fmt):
                fmt = getattr(FMTS, fmt)
        if d is None:
            self.time = dt.now()
        elif isinstance(d, str):
            from dateutil.parser import parse
            self.time = parse(d)
        elif isinstance(d, dt):
            self.time = d
        if now.pytz:
            self.time = self.time.astimezone(now.UTC)
        else:
            # timezone will be local, not UTC; don't imply UTC in format
            stderr.write('pytz not installed; now()/today() will be based on OS TZ, which may not be UTC\n')
            assert fmt[-1] == 'Z'
            fmt = fmt[:-1]

        self.fmt = fmt

    def __call__(self, fmt): return self.time.strftime(fmt)
    def __getattr__(self, k): return getattr(self.time, k)
    def __str__(self): return self.time.strftime(self.fmt)
    def __repr__(self): return str(self)
    def __int__(self): return int((self.time - now.tz(now.EPOCH)).total_seconds())
    def __float__(self): return (self.time - now.tz(now.EPOCH)).total_seconds()

    @property
    def s(self): return int(self)

    @property
    def ms(self): return int(float(self) * 1e3)

    @property
    def μs(self): return int(float(self) * 1e6)

    @property
    def us(self): return int(float(self) * 1e6)

    @staticmethod
    def to_dt(s):
        from dateutil.parser import parse
        parsed = parse(s)
        if not parsed.tzinfo:
            parsed = parsed.replace(tzinfo=now.UTC)
        return parsed

    @staticmethod
    def fromtimestamp(s, tz=UTC): return dt.fromtimestamp(s, tz=tz)

    @staticmethod
    def from_s(s): return now.fromtimestamp(s)

    @staticmethod
    def from_ms(ms): return now.fromtimestamp(ms / 1e3)

    @staticmethod
    def from_μs(us): return now.fromtimestamp(us / 1e6)

    @staticmethod
    def from_us(us): return dt.fromtimestamp(us / 1e6, tz=now.UTC)


class today(now):
    FMTS = o(
        iso='%Y-%m-%d',
        short='%Y%m%d',
    )



from os import cpu_count
from typing import Callable


def parallel(elems, fn: Callable, n_jobs: int = 0):
    if n_jobs == 1:
        return [ fn(elem) for elem in elems ]
    try:
        from joblib import Parallel, delayed
        if not n_jobs:
            n_jobs = cpu_count()
        p = Parallel(n_jobs=n_jobs)
        return p(delayed(fn)(elem) for elem in elems)
    except ImportError:
        return [ fn(elem) for elem in elems ]



# Optional-import helper
from .imports import _try

from os import environ as env

# Import most of the Python standard library
with _try: from stdlb import *

# ### Date/Time
with _try: from dateutil.parser import parse
with _try: from pytz import UTC
from .time import now, today

# ### Jupyter
with _try: from IPython.display import HTML, Image, Markdown, display

# Import other utilities from this repo:

from .backoff import backoff
from .bases import b62, b64, b90
from .hash import hash_file, HashName
from .path import mkdir, mkpar
from os import path
from .tmpdir import tmpdir

from . import process
from .process import *

from . import fn
from .fn import decos, args

with _try:
    from . import pnds
    from .pnds import *

from .cd import cd

from . import docker
from .docker.dsl import *

from .o import o
from .use import use
with _try:
    from .ym import YM
    from .ymd import YMD

with _try:
    from git import Git, Repo
    from .git import make_repo

from .git import github

from .context import *

with _try:
    from .collections import coerce, singleton, only, is_subsequence

from .defaultdict import DefaultDict, Unset

from .test import parametrize, raises

# ## Optional Modules

# joblib: easy parallelization
with _try:
    from joblib import Parallel, delayed
from .parallel import parallel

try:
    from . import plots
    from .plots import plot
    import plotly.express as px
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
except ImportError:
    pass

# yaml
with _try:
    import yaml
    # Fix a bad default in PyYAML (cf. https://github.com/yaml/pyyaml/issues/110)
    from functools import partial
    yaml.dump = partial(yaml.dump, sort_keys=False)
    yaml.safe_dump = partial(yaml.safe_dump, sort_keys=False)

# requests
with _try:
    from requests import \
        get as   GET, \
        post as  POST, \
        put as   PUT, \
        patch as PATCH

# ## PyData / Scientific Python

with _try:
    import numpy as np
    from numpy import concatenate, array, ndarray, matrix, nan

with _try: import seaborn as sns
with _try: import matplotlib.pyplot as plt
with _try: from scipy.sparse import spmatrix, coo_matrix, csr_matrix, csc_matrix

from .version import git_version, pkg_version

__version__ = pkg_version('utz')




import re
from re import escape, finditer


def split(s, ch, max: int = 0):
    """Split ``s`` on character ``ch``, unescaping backslash-escaped instances of ``ch`` (and
    unescaping double-backslashes to single backslashes).

    If ``max`` is passed, ``max`` groups will be extracted, and a final ``(max+1)``-th group will
    be appended to the returned list, containing the remainder of ``s`` (with no unescaping
    performed).
    """
    idx = 0
    strs = []
    n = 0
    nxt = ''
    escaped_ch = escape(ch)
    rgx = r'(?:\\+%s?|%s)' % (escaped_ch, escaped_ch)
    for m in finditer(rgx, s):
        start, end = m.span()
        nxt += s[idx:start]
        idx = start
        span = s[start:end]
        while span.startswith(r'\\'):
            nxt += '\\'
            span = span[2:]
            idx += 2
        if span == '\\' + ch:
            nxt += ch
            idx += 2
        elif span == ch:
            strs.append(nxt)
            nxt = ''
            idx += 1
            n += 1
            if max == n:
                break
        elif span:
            raise RuntimeError(f'Unexpected end to span {s[start:end]}: {span}; {s=}, {idx=}, {m=}')
        assert idx == end
    nxt += s[idx:]
    strs.append(nxt)
    return strs


def join(strs, ch: str, max: int = 0):
    """Escape and ``ch``-join a list of strings; inverse of ``split`` above.

    In the returned string:
    - backslashes are escaped (turned into double-backslashes)
    - instances of `ch` are escaped (preceded by a backslash)

    If ``max`` is passed and ``len(strs) == max + 1``, the final element of ``strs`` is taken to to
    have been a raw group from a corresponding call to split(…, max=…), and is directly appended to
    the final string (skipping the escaping steps above).
    """
    escaped_ch = escape(ch)

    def esc(s):
        s = re.sub(r'\\', r'\\\\', s)
        s = re.sub(escaped_ch, '\\' + escaped_ch, s)
        return s

    if max and len(strs) > max:
        if len(strs) > max + 1:
            raise ValueError(f'{len(strs)} strs > {max}+1')
        return ch.join(esc(s) for s in strs[:max]) + ch + strs[max]
    else:
        return ch.join(esc(s) for s in strs)



from dataclasses import fields
from typing import get_args, get_origin, Union


def from_dict(cls, v):
    """Parse a dataclass instance from a dictionary.

    Recursively parses instance-vars' types.
    """
    for typ in (int, float, str, bool, type(None)):
        if cls is typ:
            if type(v) is not typ:
                raise ValueError(f"{v} ({type(v).__name__}) is not of expected type {typ.__name__}")
            return v

    if get_origin(cls) is list:
        [elem_cls] = get_args(cls)
        if type(v) is not list:
            raise ValueError(f"Non-list class: {v} ({type(v)})")
        return [
            from_dict(elem_cls, _v)
            for _v in v
        ]
    elif get_origin(cls) is dict:
        if not isinstance(v, dict):
            raise ValueError(f"Expected {cls}, got {v})")
        [key_cls, value_cls] = get_args(cls)
        return {
            from_dict(key_cls, k): from_dict(value_cls, _v)
            for k, _v in v.items()
        }
    elif get_origin(cls) is Union:
        args = get_args(cls)
        for arg in args:
            try:
                return from_dict(arg, v)
            except (TypeError, ValueError) as e:
                pass
        raise ValueError(f"Invalid value '{v}' (expected one of {[ c.__name__ for c in get_args(cls) ]})")

    try:
        fieldtypes = { f.name: f.type for f in fields(cls) }
    except TypeError:
        raise TypeError(f"`fields` must be called with a dataclass type or instance, not {getattr(cls, '__name__', cls)}")

    if not isinstance(v, dict):
        raise ValueError(f"Expected dict (representing a {cls.__name__}), got {v}")

    kwargs = {}
    for k, _v in v.items():
        fieldtype = fieldtypes[k]
        kwargs[k] = from_dict(fieldtype, _v)
    return cls(**kwargs)



#!/usr/bin/env python

from pathlib import Path
from PyPDF2 import PdfFileWriter, PdfFileReader
from io import BytesIO
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import letter
from subprocess import check_output


def maybe_open_pdf(path, open_pdf):
    if open_pdf:
        check_output(['open', path])


def write(path, out, *args, page=0, font_size=None, open_pdf=False):
    path = str(path)
    out = str(out)
    page_idx = page

    if isinstance(args[0], int) and isinstance(args[1], int) and isinstance(args[2], str):
        args = [args]
    
    packet = BytesIO()
    # create a new PDF with Reportlab
    can = canvas.Canvas(packet, pagesize=letter)
    if font_size:
        can.setFontSize(font_size)
    
    for x, y, s in args:
        can.drawString(x, y, s)
    
    can.save()

    # move to the beginning of the StringIO buffer
    packet.seek(0)
    new_pdf = PdfFileReader(packet)
    # read your existing PDF
    existing_pdf = PdfFileReader(open(path, "rb"))
    output = PdfFileWriter()
    
    # add the "watermark" (which is the new pdf) on the existing page
    page = existing_pdf.getPage(page_idx)
    page.mergePage(new_pdf.getPage(page_idx))
    output.addPage(page)

    # finally, write "output" to a real file
    with open(out, "wb") as f:
        output.write(f)

    maybe_open_pdf(out, open_pdf)


def cat(inputs, output, open_pdf=False):
    """https://stackoverflow.com/a/3444735"""
    input_streams = []
    try:
        # First open all the files, then produce the output file, and
        # finally close the input files. This is necessary because
        # the data isn't read from the input files until the write
        # operation. Thanks to
        # https://stackoverflow.com/questions/6773631/problem-with-closing-python-pypdf-writing-getting-a-valueerror-i-o-operation/6773733#6773733
        for input in inputs:
            input = Path(input)
            input_streams.append(input.open('rb'))
        writer = PdfFileWriter()
        for reader in map(PdfFileReader, input_streams):
            for n in range(reader.getNumPages()):
                writer.addPage(reader.getPage(n))

        if isinstance(output, Path):
            with output.open('wb') as f:
                writer.write(f)
        elif isinstance(output, str):
            with open(output, 'wb') as f:
                writer.write(f)
        else:
            writer.write(output)
    finally:
        for f in input_streams:
            f.close()
    
    maybe_open_pdf(output, open_pdf)



from inspect import getfullargspec
from typing import Callable, Sequence, Union

Deco = Callable[[Callable], Callable]


def decos(*args: Union[Deco, Sequence[Deco]]):
    """Compose decorators."""
    decos = [ deco for decos in args for deco in decos ]

    def _fn(fn):
        for deco in reversed(decos):
            fn = deco(fn)
        return fn

    return _fn


def args(fn, kwargs):
    """Filter kwargs to match function signature."""
    spec = getfullargspec(fn)
    args = spec.args
    return { k: v for k, v in kwargs.items() if k in args }


def call(fn, *_args, **kwargs):
    """Call a function with only the kwargs that it is able to receive."""
    return fn(*_args, args(fn, **kwargs))



import re

CAPS_RGX = re.compile(r'(?<!^)(?=[A-Z])')


def dash_case(orig):
    if '_' in orig:
        return re.sub("_", "-", orig)
    else:
        return CAPS_RGX.sub('-', orig).lower()


def snake_case(orig):
    if '-' in orig:
        return re.sub("-", "_", orig)
    else:
        return CAPS_RGX.sub('_', orig).lower()


def camel_case(s):
    def upper_first(pc):
        return pc[0].upper() + pc[1:]
    return ''.join([ upper_first(pc) for pc in re.split('[-_:]', s) ])



from importlib.metadata import version, PackageNotFoundError
from re import fullmatch
from subprocess import CalledProcessError

from .process import line


VERSION_TAG_REGEX = r"v?(?P<version>(?P<base>\d+\.\d+\.\d+)(?P<rc>(?:r|c|rc|a|b)\d+)?(?:-(?P<commits_ahead>\d+)-g(?P<sha>[0-9a-f]{6,}))?)"


def git_version():
    """Infer version from git tag of the form "v_._._"; otherwise, return current Git commit SHA"""
    try:
        tag = line('git', 'describe', '--tags', 'HEAD')
        m = fullmatch(VERSION_TAG_REGEX, tag)
        if not m:
            raise ValueError('Unrecognized tag: %s' % tag)
        version = m['version']
    except CalledProcessError:
        version = line('git', 'log', '--format=%h', '-n1')

    return version


def pkg_version(name=None):
    try:
        return version(name)
    except PackageNotFoundError:
        return git_version()



#!/usr/bin/env python

import dask.dataframe as dd
import pandas as pd
from os import cpu_count


def to_sql(
    table, name, db_path, force=False,
):
    if_exists = 'replace' if force else 'fail'
    table.to_sql(name, db_path, if_exists=if_exists)
    if isinstance(table, dd.DataFrame):
        divisions = pd.Series(table.divisions, name='divisions')
        divisions.to_sql(f'{name}/divisions', db_path, if_exists=if_exists)


def from_sql(
    name, db_path, index_col=None, dask=True,
):
    if index_col is None:
        index_col = 'index'

    if dask:
        divisions = pd.read_sql_table(f'{name}/divisions', db_path, index_col='index').divisions.tolist()
        table = dd.read_sql_table(name, db_path, index_col=index_col, divisions=divisions)
    else:
        table = pd.read_sql_table(name, db_path, index_col=index_col)

    return table


def table_to_sql(
    table_name,
    db_path,
    index_col=None,
    dask=True,
    npartitions=-1,
    table_path=None,
    force_db_refresh=False,
    **kwargs
):
    from sqlalchemy import create_engine
    engine = create_engine(db_path)

    if npartitions is not None:
        if npartitions <= 0:
            npartitions = cpu_count()

    if force_db_refresh or not engine.has_table(table_name):
        if not table_path:
            raise ValueError(f'table_path required in order to compute table {table_name}')
        print(f'Creating db table: {table_name}')
        if_exists = ('replace' if force_db_refresh else 'fail')
        if dask:
            csv = dd.read_csv(table_path, **kwargs)
            if npartitions is not None:
                csv = csv.repartition(npartitions=npartitions)
            if index_col is None:
                # Force Dask to get a cross-partition default integer autoinc index
                csv = csv.reset_index().set_index('index')
        else:
            csv = pd.read_csv(table_path, **kwargs)
        
        if index_col is not None:
            csv = csv.set_index(index_col)

        to_sql(csv, table_name, db_path, force=force_db_refresh)

    return from_sql(table_name, db_path, index_col=index_col, dask=dask)



import re
from dataclasses import dataclass
from datetime import datetime as dt, date, timedelta
from functools import wraps
from typing import Union, Generator
from utz.imports import _try


# Types that can be passed to the Month constructor
Monthy = Union['YMD', str, int, None]


@dataclass(init=False, order=True, eq=True, unsafe_hash=True)
class YMD:
    y: int
    m: int
    d: int

    RGX = re.compile(r'(?P<year>\d{4})(?:-?(?P<month>\d\d)(?:-?(?P<day>\d\d))?)?')

    def _init_from_str(self, arg):
        m = self.RGX.fullmatch(arg)
        if not m:
            raise ValueError('Invalid month string: %s' % arg)
        year = int(m['year'])
        month = int(m['month']) if m['month'] else 1
        day = int(m['day']) if m['day'] else 1
        if month > 12:
            raise ValueError(f"Invalid month {month} ({arg})")
        self.__init__(year, month, day)

    def _verify(self):
        if not isinstance(self.y, int):
            raise ValueError('Year %s must be int, not %s' % (str(self.y), type(self.y)))
        if not isinstance(self.m, int):
            raise ValueError('Month %s must be int, not %s' % (str(self.m), type(self.m)))
        if not isinstance(self.d, int):
            raise ValueError('Day %s must be int, not %s' % (str(self.d), type(self.d)))

    def _init_now(self):
        now = dt.now()
        self.y = now.year
        self.m = now.month
        self.d = now.day

    def __init__(self, *args, **kwargs):
        if kwargs:
            if args:
                raise ValueError(f'Pass args xor kwargs: {args}, {kwargs}')
            keys = list(kwargs.keys())
            if keys == ['y', 'm', 'd']:
                self.y = kwargs['y']
                self.m = kwargs['m']
                self.d = kwargs['d']
            else:
                raise ValueError(f"Unrecognized kwargs: {kwargs}")
        elif len(args) == 3:
            self.y, self.m, self.d = int(args[0]), int(args[1]), int(args[2])
            self._verify()
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, str):
                self._init_from_str(arg)
            elif isinstance(arg, int):
                self._init_from_str(str(arg))
            elif hasattr(arg, 'year') and hasattr(arg, 'month') and hasattr(arg, 'day'):
                self.y = arg.year
                self.m = arg.month
                self.d = arg.day
                self._verify()
            elif arg is None:
                self._init_now()
            elif 'year' in arg and 'month' in arg and 'day' in arg:
                self.y = int(arg['year'])
                self.m = int(arg['month'])
                self.d = int(arg['day'])
                self._verify()
            else:
                raise ValueError('Unrecognized argument: %s' % str(arg))
        elif not args:
            self._init_now()
        else:
            raise ValueError('Unrecognized arguments: %s' % str(args))

    @property
    def year(self):
        return self.y

    @property
    def month(self):
        return self.m

    @property
    def day(self):
        return self.d

    def str(self, sep=''):
        return '%d%s%02d%s%02d' % (self.y, sep, self.m, sep, self.d)

    def __str__(self):
        return self.str()

    def __int__(self):
        return int(str(self))

    def format(self, url, **kwargs):
        return url.format(ymd=str(self), y=str(self.y), m=str(self.m), d=str(self.d), **kwargs)

    @property
    def dt(self):
        import pandas as pd
        return pd.to_datetime('%d-%02d-%02d' % (self.y, self.m, self.d))

    @property
    def date(self) -> date:
        return self.dt.date()

    def __add__(self, n: int) -> 'YMD':
        if not isinstance(n, int):
            raise ValueError('%s: can only add an integer to a Month, not %s: %s' % (str(self), str(type(n)), str(n)))
        return YMD(self.date + timedelta(days=n))

    def __sub__(self, n: int) -> 'YMD':
        if not isinstance(n, int):
            raise ValueError('%s: can only add an integer to a Month, not %s: %s' % (str(self), str(type(n)), str(n)))
        return YMD(self.date - timedelta(days=n))

    def until(self, end: 'YMD' = None, step: int = 1) -> Generator['YMD', None, None]:
        cur: YMD = YMD(self)
        while end is None \
                or (step > 0 and cur < end) \
                or (step < 0 and cur > end):
            yield cur
            cur = cur + step


def dates(*flags, default_start=None, default_end=None, help=None):
    if not flags:
        flags = ('-d', '--dates')

    from click import option
    def _dates(fn):
        @option(*flags, help=help)
        @wraps(fn)
        def _fn(*args, dates=None, **kwargs):
            if dates:
                pcs = dates.split('-')
                if len(pcs) == 2:
                    [ start, end ] = pcs
                    start = YMD(start) if start else default_start
                    end = YMD(end) if end else default_end
                elif len(pcs) == 1:
                    [ym] = pcs
                    ym = YMD(ym)
                    start = ym
                    end = ym + 1
                else:
                    raise ValueError(f"Unrecognized {'/'.join(flags)}: {dates}")
            else:
                start, end = default_start, default_end
            fn(*args, start=start, end=end, **kwargs)

        return _fn

    return _dates



from os import makedirs
from os.path import dirname
import pathlib
from typing import Union

Path = Union[str, pathlib.Path]


def mkdir(path: Path, mode=0o777, exist_ok=True) -> Path:
    """Wrapper for ``makedirs``, accepting ``str | Path``, and setting ``exist_ok=True`` by default."""
    makedirs(str(path), mode=mode, exist_ok=exist_ok)
    return path


def mkpar(path: Path, *args, **kwargs) -> Path:
    """Create parent directories for a path, if they do not exist."""
    if isinstance(path, Path):
        mkdir(path.parent, *args, **kwargs)
    else:
        mkdir(dirname(path))
    return path
